//------------------------------------------------------------------------------
//
//  npplot.c
//  npe
//
//  Created by Shane Saxon on 2/19/10.
//  Copyright Saxon Digital 2010. All rights reserved.
//
//------------------------------------------------------------------------------


#include "npplot.h"

#include <stdlib.h>

#include "../npctrl.h"


void InitPlotDefault (NPplotPtr plot);
void InitPlotCamera (NPplotPtr plot);
void InitPlotSurface (NPplotPtr plot);
void InitPlotPoints (NPplotPtr plot);
void InitPlotPin (NPplotPtr plot);
void InitPlotVideo (NPplotPtr plot);
void InitPlotGrid (NPplotPtr plot);


//------------------------------------------------------------------------------
void npInitPlot (void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < kPlotMax; i++)
		data->map.plot[i] = NULL;
}


//------------------------------------------------------------------------------
void npClosePlot(void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	for (i=0; i < data->map.plotCount; i++)
		npPlotDelete (data->map.plot[i], dataRef);
}


// should make this static and add locking for delete to be thread safe,      debug zz
//------------------------------------------------------------------------------
void* npPlotNew (int plotType, void* dataRef)
{
	int plotIndex = 0;
	NPplotPtr plot = NULL;

	pData data = (pData) dataRef;

	//return if at the plot max
	if ( data->map.plotCount == kPlotMax )
		return NULL;							// add error plotmax , debug zz

	//init the plot
	plot = (NPplotPtr) malloc (sizeof(NPplot));

	InitPlotDefault (plot);	

	plot->type = plotType;

	switch (plotType)
	{
		case kPlotCamera :	InitPlotCamera (plot); break;
		case kPlotGrid :	InitPlotGrid (plot); break;
		case kPlotVideo :	InitPlotVideo (plot); break;
		case kPlotSurface :	InitPlotSurface (plot); break;
		case kPlotPoints :	InitPlotPoints (plot); break;
		case kPlotPin :		InitPlotPin (plot); break;
		case kPlotPinTorus :
			InitPlotPin (plot);
			plot->surfaceType = kSurfaceTorus;
			return (void*)plot;
			break;

		default : 
			break;
	}

	//add the plot to the list, replace with a double linked list later, debug, zz

	npPlotNewChild (plot, dataRef);

	data->map.plot[data->map.plotCount] = plot;
	data->map.plotIndex = data->map.plotCount;
	data->map.plotCount++;

	data->map.plotChildPtr = plot;
	data->map.branchLevel = 0;

	return (void*)plot;
}

					// add a primary torus child to the trunk,
					//					make this part of the npPlotNew.. debug, zz
void* npPlotNewChild (NPplotPtr plot, void* dataRef)
{
	NPplotPtr plotChild = NULL;


	// create and add child to level 1
	if (plot->childCount >= kPlotChildMax)
		return NULL;

	plotChild = npPlotNew (kPlotPinTorus, dataRef);
					
	plotChild->parent = plot;
	plotChild->branchLevel = plot->branchLevel + 1;

	// distribute the children at 25 deg spacing... is 0.44 radians
	plotChild->rotateRad.x = 0.44f * (GLfloat)plot->childCount;

	// attach child to parent
	plot->child[plot->childCount] = plotChild;
	plot->childCount++;
	plot->childIndex = plot->childCount - 1;

	return (void*)plotChild;
}

															// add travers tree delete...
//------------------------------------------------------------------------------
void npPlotDelete (NPplotPtr plot, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;
	NPplotPtr parent = NULL;

	if (data->map.plotIndex < 3 || plot == NULL)	//do not delete the root grid or camera
		return;

	if (plot->branchLevel > 0)
	{
		parent = plot->parent;

		// don't delete the last torus...			// fix thi
		if (plot->branchLevel == 1)
		{
			data->map.branchLevel = 0;
			data->map.plotChildPtr = parent;
			return;
		}

		parent->childCount--;

		// compact the gap, optimize with block copy later...		zz
		for (i = parent->childIndex; i < parent->childCount; i++)
			parent->child[i] = parent->child[i+1];

		parent->child[parent->childCount] = NULL;

		if (parent->childIndex >= parent->childCount)
			parent->childIndex = parent->childCount - 1;

		if (parent->childCount == 0)
		{
			data->map.branchLevel = parent->branchLevel;
			data->map.plotChildPtr = parent;
		}
		else
			data->map.plotChildPtr = parent->child[parent->childIndex];
	}
	else
	{
		data->map.plotCount--;

		// compact the gap, optimize with block copy later...		zz
		for (i = data->map.plotIndex; i < data->map.plotCount; i++)
			data->map.plot[i] = data->map.plot[i+1];

		data->map.plot[data->map.plotCount] = NULL;

		if (data->map.plotIndex >= data->map.plotCount)
			data->map.plotIndex = data->map.plotCount - 1;

		switch (plot->type)
		{
			case kPlotCamera :	free ((NPcameraPtr) plot->data); break;
			case kPlotVideo :	free ((NPvideoPtr) plot->data); break;
			case kPlotSurface : free ((NPsurfacePtr) plot->data); break;
			case kPlotPoints :	free ((NPpointsPtr) plot->data); break;
			case kPlotPin :		free ((NPpinPtr) plot->data); break;
			case kPlotGrid :	free ((NPgridPtr) plot->data); break;
			default : 
				break;
		}
		data->map.plotChildPtr = data->map.plot[data->map.plotIndex];
	}
	free (plot);
}


//------------------------------------------------------------------------------
void InitPlotDefault (NPplotPtr plot)
{
	int i = 0;

	plot->type			= kPlotDefault;		//plot type, point, pin, cam, video
	plot->data			= NULL;				//plot type specific data

	plot->selected		= 0;				//defualt is not selected

	plot->parent		= NULL;				//parent plot, binary tree of plots
	plot->branchLevel	= 0;			//0 is the trunk pin (ice-cream cone)

	for (i=0; i < kPlotChildMax; i++)
		plot->child[i] = NULL;

	plot->childCount	= 0;				//no children by default
	plot->childIndex	= 0;

	plot->channel.x		= kChannelNull;		//channel to use for x,y,z points
	plot->channel.y		= kChannelNull;
	plot->channel.z		= kChannelNull;
	
	plot->channelIndex.x = 0;				//last updated channel index
	plot->channelIndex.y = 0;
	plot->channelIndex.z = 0;

	plot->averageType.x	= kAverageNull;		//averaging type applied to data
	plot->averageType.y	= kAverageNull;
	plot->averageType.z	= kAverageNull;
	
	plot->sampleInterval.x = 1;				//the interval to be averaged
	plot->sampleInterval.y = 1;
	plot->sampleInterval.z = 1;

	plot->rotate.angle	= 0.0f;				//orientation vector
	plot->rotate.x		= 0.0f;
	plot->rotate.y		= 0.0f;
	plot->rotate.z		= 1.0f;
	
	plot->scale.x		= 1.0f;				//scale plot XYZ
	plot->scale.y		= 1.0f;
	plot->scale.z		= 1.0f;
	
	plot->translate.x	= 0.0f;				//location relative to plotOrigin
	plot->translate.y	= 0.0f;
	plot->translate.z	= 0.0f;

	plot->plotOrigin.x	= 0.0f;				//origin relative to parent origin
	plot->plotOrigin.y	= 0.0f;
	plot->plotOrigin.z	= 0.0f;
	
	plot->rotateRate.x	= 0.0f;
	plot->rotateRate.y	= 0.0f;
	plot->rotateRate.z	= 0.0f;

	plot->rotateRad.x	= 0.0f;
	plot->rotateRad.y	= 0.0f;
	plot->rotateRad.z	= 0.0f;
	
	plot->scaleRate.x	= 0.0f;				//scale rate 
	plot->scaleRate.y	= 0.0f;
	plot->scaleRate.z	= 0.0f;

	plot->translateRate.x = 0.0f;			//pans the location of the plot, pronunciar en Espanol, lol
	plot->translateRate.y = 0.0f;
	plot->translateRate.z = 0.0f;

	plot->translateVect.x = 0.0f;			//pans the location of the plot, pronunciar en Espanol, lol
	plot->translateVect.y = 0.0f;
	plot->translateVect.z = 0.0f;
	
	plot->surfaceType	= kSurfaceNull;		//FFT layers, mesh, sphere...
	plot->shaderType	= kShadingFlat;		//wire, flat, gouraud, phong

	plot->lineWidth		= 0.0f;				//draws lines, zero for off
	plot->pointSize		= 0.0f;				//draws dots, zero for off
	plot->circle		= 0.0f;				//draw circle at current data point

	plot->colorIndex	= 0;
	
	plot->color.r		= 50;				//default color assigned to new data
	plot->color.g		= 101;
	plot->color.b		= 101;
	plot->color.a		= 255;				//fully opaque

	plot->colorFade		= false;

	plot->hide			= false;			//hide the plot, data stays active
	plot->freeze		= false;			//freezes data and physics updates
	plot->center		= false;			//centers to plotOrigin
	plot->centerPerGrid	= false;			//centers in gridSpace increments

	plot->autoZoom.x	= false;			//scales plot to fit screen
	plot->autoZoom.y	= false;
	plot->autoZoom.z	= false;

	plot->scroll		= false;			//scroll right to left
	
	plot->triggerOnHi.x	= false;			//triggers are multipurpose
	plot->triggerOnHi.y	= false;
	plot->triggerOnHi.x	= false;

	plot->triggerOnLow.x = false;
	plot->triggerOnLow.y = false;
	plot->triggerOnLow.z = false;
	
	plot->triggerLevelHi.x = 0.0f;
	plot->triggerLevelHi.y = 0.0f;
	plot->triggerLevelHi.z = 0.0f;

	plot->triggerLevelLow.x = 0.0f;
	plot->triggerLevelLow.y = 0.0f;
	plot->triggerLevelLow.z = 0.0f;

	plot->triggerCommand = kCtrlCommandNull;	//execute ctrlCommand upon trigger
	
	plot->proximity.x	= false;			//used for collision engine 
	plot->proximity.y	= false;
	plot->proximity.z	= false;
	
	plot->proximityRadius.x = 0.0f;
	plot->proximityRadius.y = 0.0f;
	plot->proximityRadius.z = 0.0f;

	plot->drawCurrentValue.x = false;		//draw current values as text
	plot->drawCurrentValue.x = false;
	plot->drawCurrentValue.x = false;

	plot->drawTranslate.x = false;			//draw translate as text
	plot->drawTranslate.x = false;
	plot->drawTranslate.x = false;
	
	plot->drawScale.x	= false;			//draw scale as text
	plot->drawScale.x	= false;
	plot->drawScale.x	= false;
	
	plot->drawRotate.x	= false;			//draw rotation as text
	plot->drawRotate.y	= false;
	plot->drawRotate.z	= false;
}


// cameras can be GL scene cameras or real-world cameras
// video is attached as a daughter of the recPlot
//-----------------------------------------------------------------------------
void InitPlotCamera (NPplotPtr plot)
{
	NPcameraPtr data = (NPcameraPtr) malloc (sizeof(NPcamera));

	plot->data  = (void*) data;

	plot->type			= kPlotCamera;

	data->textureID		= 0;					//texture ID in GL

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->fov			= 35.0f;				//35mm, 70mm...
	data->clipNear		= 1.0f;					//clipping planes
	data->clipFar		= 1000.0f;

	data->aperture		= 1.0f;					//F stop
	data->exposure		= 0.01f;				//in seconds
	data->sensorType	= kCameraSensorNull;	//3CCD, Debayer pattern...
}


//-----------------------------------------------------------------------------
void InitPlotVideo (NPplotPtr plot)
{
	NPvideoPtr data = (NPvideoPtr) malloc (sizeof(NPvideo));

	plot->data  = (void*) data;

	plot->type			= kPlotVideo;

	data->textureID		= 0;

	data->format		= kVideoFormat720p;
	data->interlaced	= kVideoFieldNull;		//progressive scan, no fields
	data->stereo		= kVideoStereoNull;		//not stereoscopic
	data->aspectRatio	= 1.77778f;				//1.77778 is approx 16:9
	data->colorSpace	= kColorSpaceRGB_8;		//8bit RGB is what GPU's output
	data->fps			= 59.97f;
	data->width			= 1280;
	data->height		= 720;

	data->bias.r		= 0.0f;				//brightness shift, 0.0 no change
	data->bias.g		= 0.0f;
	data->bias.b		= 0.0f;
	data->bias.a		= 0.0f;
	data->gain.r		= 1.0f;				//color multiplier, 1.0 no change
	data->gain.g		= 1.0f;
	data->gain.b		= 1.0f;
	data->gain.a		= 1.0f;
}


//------------------------------------------------------------------------------
void InitPlotSurface (NPplotPtr plot)
{
	NPsurfacePtr data = (NPsurfacePtr) malloc (sizeof(NPsurface));

	plot->data  = (void*) data;

	plot->type			= kPlotSurface;

	data->textureID		= 0;

	data->vertexArray	= NULL;
	data->normalArray	= NULL;
	data->colorArray	= NULL;
	data->vertexCount	= 0;
	data->normalCount	= 0;
	data->arrayIndex	= 0;	
}


//-----------------------------------------------------------------------------
void InitPlotPoints (NPplotPtr plot)
{
	NPpointsPtr data = (NPpointsPtr) malloc (sizeof(NPpoints));

	plot->data  = (void*) data;
	
	plot->type			= kPlotPoints;

	data->vertexArray	= NULL;				//the array of data points
	data->colorArray	= NULL;				//each point gets color assigned
	data->vertexCount	= 0;				//number of points in the plot
	data->arrayIndex	= 0;				//current vertex index in the array
}


//-----------------------------------------------------------------------------
void InitPlotPin (NPplotPtr plot)
{
	NPpinPtr data = (NPpinPtr) malloc (sizeof(NPpin));

	plot->data  = (void*) data;

	plot->type			= kPlotPin;

	plot->surfaceType	= kSurfacePin;		//ice-cream cone shape

	data->textureID		= 0;				//0 for no texture assigned

	data->innerRadius	= 0.15f;			//inner radius of the toroid
	data->outerRadius	= 1.5f;			//outer radius of the toroid

	// 4 * outerRadius / innerRadius -> int
	data->radiusRatioIndex = 40;			//sorts pre-loaded GPU toriods
	data->screenSizeIndex = 8;				//same as above for detail level

	data->slices		= 7;				//calculated from screenSizeIndex
	data->stacks		= 16;

	data->translateTex.x = 0.0f;			//position offset of texture map
	data->translateTex.y = 0.0f;
	data->translateTex.z = 0.0f;

	data->scaleTex.x	= 1.0f;
	data->scaleTex.y	= 1.0f;
	data->scaleTex.z	= 1.0f;

	data->rotateTex.angle = 0.0f;			//orientation of texture map
	data->rotateTex.x	= 0.0f;
	data->rotateTex.y	= 0.0f;
	data->rotateTex.z	= 0.0f;

//	plot->translate.x += 10.0f;				// check if trunk then do this debug, zz
}


//------------------------------------------------------------------------------
void InitPlotGrid (NPplotPtr plot)
{
	NPgridPtr data = (NPgridPtr) malloc (sizeof(NPgrid));

	plot->data  = (void*) data;

	plot->type			= kPlotGrid;

	plot->colorIndex	= 14;
	plot->color.r		= 152;
	plot->color.g		= 0;
	plot->color.b		= 0;
	plot->color.a		= 255;

	data->segments.x	= 10;
	data->segments.y	= 10;
	data->segments.z	= 1;
	data->size.x		= 10.0f;
	data->size.y		= 10.0f;
	data->size.z		= 10.0f;

	data->overlay		= false;
}

