//------------------------------------------------------------------------------
//
//  npctrl.c
//  npe
//
//  Created by Shane Saxon on 2/19/10.
//  Copyright Saxon Digital 2010. All rights reserved.
//
//------------------------------------------------------------------------------

#include "npdata.h"
#include "npio.h"
#include "npctrl.h"

#include "io/npglut.h"
#include "io/npgl.h"
#include "ctrl/npengine.h"

void npCtrlVector (int command, void* dataRef);
void npCtrlSelect (int command, void* dataRef);
void npCtrlProperty (int command, void* dataRef);
void npCtrlGlobal (int command, void* dataRef);

//------------------------------------------------------------------------------
void npInitCtrl (void* dataRef)
{
	npInitEngine (dataRef);
}

//------------------------------------------------------------------------------
void npCloseCtrl (void* dataRef)
{
	npCloseEngine (dataRef);
}

//------------------------------------------------------------------------------
int npMainAppLoop (void* dataRef)
{
	int err = 0;

#ifdef _LINUX_
//	MainAppLoopLinux();
#endif

#ifdef _MSW_
//	MainAppLoopMSW();
	npInitGlut (0, 0, dataRef);
	npMainAppLoopGlut (dataRef);
#endif

#ifdef _OSX_
	//	MainAppLoopOSX();
#endif


	return err;
}


//------------------------------------------------------------------------------
void npCtrlCommand (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	NPplotPtr plot = data->map.plot[data->map.plotIndex];

	switch (command)
	{
		//translate the selected objects relative to the camera
		case kGraphXdecrease : npCtrlVector (command, dataRef); break;
		case kGraphXdecreaseOff : npCtrlVector (command, dataRef); break;
		case kGraphXincrease : npCtrlVector (command, dataRef); break;
		case kGraphXincreaseOff : npCtrlVector (command, dataRef); break;

		case kGraphYdecrease : npCtrlVector (command, dataRef); break;
		case kGraphYdecreaseOff : npCtrlVector (command, dataRef); break;
		case kGraphYincrease : npCtrlVector (command, dataRef); break;
		case kGraphYincreaseOff : npCtrlVector (command, dataRef); break;

		case kGraphZdecrease : npCtrlVector (command, dataRef); break;
		case kGraphZdecreaseOff : npCtrlVector (command, dataRef); break;
		case kGraphZincrease : npCtrlVector (command, dataRef); break;
		case kGraphZincreaseOff : npCtrlVector (command, dataRef); break;

		case kGraphRotateLeft : npCtrlVector (command, dataRef); break;
		case kGraphRotateLeftOff : npCtrlVector (command, dataRef); break;
		case kGraphRotateRight : npCtrlVector (command, dataRef); break;
		case kGraphRotateRightOff : npCtrlVector (command, dataRef); break;

		case kGraphRotateUp : npCtrlVector (command, dataRef); break;
		case kGraphRotateUpOff : npCtrlVector (command, dataRef); break;
		case kGraphRotateDown : npCtrlVector (command, dataRef); break;
		case kGraphRotateDownOff : npCtrlVector (command, dataRef); break;

		case kGraphRotateCCW : npCtrlVector (command, dataRef); break;
		case kGraphRotateCCWOff : npCtrlVector (command, dataRef); break;
		case kGraphRotateCW : npCtrlVector (command, dataRef); break;
		case kGraphRotateCWOff : npCtrlVector (command, dataRef); break;

		case kGraphZoomOn : npCtrlVector (command, dataRef); break;
		case kGraphZoomOff : npCtrlVector (command, dataRef); break;
		
		// Selects the active axes and plots to manipulate
		case kGraphSelectAxes : npCtrlSelect (command, dataRef); break;

		case kGraphGrid : npCtrlSelect (command, dataRef); break;
		case kGraphCamera : npCtrlSelect (command, dataRef); break;

		case kGraphNext : npCtrlSelect (command, dataRef); break;
		case kGraphNextBranch : npCtrlSelect (command, dataRef); break;

		case kGraphNew : npCtrlSelect (command, dataRef); break;
		case kGraphDelete : npCtrlSelect (command, dataRef); break;

		case kGraphSelect : npCtrlSelect (command, dataRef); break;
		case kGraphSelectOne : npCtrlSelect (command, dataRef); break;
		case kGraphSelectTwo : npCtrlSelect (command, dataRef); break;
		case kGraphSelectThree : npCtrlSelect (command, dataRef); break;
		case kGraphSelectAll : npCtrlSelect (command, dataRef); break;

		// Change plot property
		case kGraphChannelUp : npCtrlProperty (command, dataRef); break;
		case kGraphChannelDown : npCtrlProperty (command, dataRef); break;

		case kGraphColorUp : npCtrlProperty (command, dataRef); break;
		case kGraphColorDown : npCtrlProperty (command, dataRef); break;
		case kGraphAlphaUp : npCtrlProperty (command, dataRef); break;
		case kGraphAlphaDown : npCtrlProperty (command, dataRef); break;
					
		case kGraphCenter : npCtrlProperty (command, dataRef); break;
		case kGraphLines : npCtrlProperty (command, dataRef); break;
		case kGraphPoints : npCtrlProperty (command, dataRef); break;
		case kGraphSurface : npCtrlProperty (command, dataRef); break;
		case kGraphMesh : npCtrlProperty (command, dataRef); break;
		case kGraphScroll : npCtrlProperty (command, dataRef); break;
		case kGraphFreeze : npCtrlProperty (command, dataRef); break;
		case kGraphHide : npCtrlProperty (command, dataRef); break;
		case kGraphClear : npCtrlProperty (command, dataRef); break;

		case kGraphSampleInterval : npCtrlProperty (command, dataRef); break;
		case kEventLowSetpoint : npCtrlProperty (command, dataRef); break;
		case kEventHighSetpoint : npCtrlProperty (command, dataRef); break;

		case kGraphDrawIndicators : npCtrlProperty (command, dataRef); break;

		// Change global property
		case kGraphMenu : npCtrlGlobal (command, dataRef); break;
		case kFullScreenExit : npCtrlGlobal (command, dataRef); break;
		case kGraphClearFrameBuffer : npCtrlGlobal (command, dataRef); break;

		case kGraphPresetOne : npDataPreset (command, dataRef); break;
		case kGraphPresetTwo : npDataPreset (command, dataRef); break;
		case kGraphPresetThree : npDataPreset (command, dataRef); break;
		case kGraphPresetFour : npDataPreset (command, dataRef); break;
		case kGraphPresetFive : npDataPreset (command, dataRef); break;
		case kGraphPresetSix : npDataPreset (command, dataRef); break;
		case kGraphPresetSeven : npDataPreset (command, dataRef); break;
		case kGraphPresetEight : npDataPreset (command, dataRef); break;

		default : break;
	}
}


// update the translateVect for selected objects
// translateVect is relative to global coordinates
// translateRate is relative to the camera
//------------------------------------------------------------------------------
void npCtrlVector (int command, void* dataRef)
{
	GLfloat speed = 1.0f;

	pData data = (pData) dataRef;

	// most commands only effect pin trunk, some effect branches...
	NPplotPtr plot = data->map.plot[data->map.plotIndex];

	// slow down movement if shift key pressed
	if (data->io.key.modShift)
		speed = 0.2f;

	switch (command)
	{
		case kGraphXdecrease : plot->translateRate.x = -0.2f * speed; break;
		case kGraphXdecreaseOff : 
			if (plot->translateRate.x < 0.0f)	//graceful multi-key behavior
				plot->translateRate.x = 0.0f; 
			break;
		case kGraphXincrease : plot->translateRate.x = 0.2f * speed; break;
		case kGraphXincreaseOff : 
			if (plot->translateRate.x > 0.0f)
				plot->translateRate.x = 0.0f; 
			break;

		case kGraphYdecrease : plot->translateRate.y = -0.2f * speed; break;
		case kGraphYdecreaseOff : 
			if (plot->translateRate.y < 0.0f)
				plot->translateRate.y = 0.0f;
			break;
		case kGraphYincrease : plot->translateRate.y = 0.2f * speed; break;
		case kGraphYincreaseOff : 
			if (plot->translateRate.y > 0.0f)
				plot->translateRate.y = 0.0f;
			break;

		case kGraphZdecrease : plot->translateRate.z = -0.2f * speed; break;
		case kGraphZdecreaseOff : 
			if (plot->translateRate.z < 0.0f)
				plot->translateRate.z = 0.0f;
			break;
		case kGraphZincrease : plot->translateRate.z = 0.2f * speed; break;
		case kGraphZincreaseOff : 
			if (plot->translateRate.z > 0.0f)
				plot->translateRate.z = 0.0f;
			break;
		
		case kGraphRotateLeft : plot->rotateRate.z = -0.005f * speed; break;
		case kGraphRotateLeftOff : 
			if (plot->rotateRate.z < 0.0f)
				plot->rotateRate.z = 0.0f;
			break;
		case kGraphRotateRight : plot->rotateRate.z = 0.005f * speed; break;
		case kGraphRotateRightOff : 
			if (plot->rotateRate.z > 0.0f)
				plot->rotateRate.z = 0.0f;
			break;

		case kGraphRotateUp : 
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			plot->rotateRate.x = 0.005f * speed; break;
		case kGraphRotateUpOff :
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			if (plot->rotateRate.x > 0.0f)
				plot->rotateRate.x = 0.0f;
			break;
		case kGraphRotateDown : 
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			plot->rotateRate.x = -0.005f * speed; break;
		case kGraphRotateDownOff :
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			if (plot->rotateRate.x < 0.0f)
				plot->rotateRate.x = 0.0f;
			break;

		case kGraphRotateCCW : plot->rotateRate.y = -0.005f * speed; break;
		case kGraphRotateCCWOff : 
			if (plot->rotateRate.y < 0.0f)
				plot->rotateRate.y = 0.0f;
			break;
		case kGraphRotateCW : plot->rotateRate.y = 0.005f * speed; break;
		case kGraphRotateCWOff : 
			if (plot->rotateRate.y > 0.0f)
				plot->rotateRate.y = 0.0f;
			break;

		case kGraphZoomOn :
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			if (data->io.key.modShift)
			{
				if (data->io.axes.x == true )
					plot->scaleRate.x = -0.003f;
				if (data->io.axes.y == true )
					plot->scaleRate.y = -0.003f;
				if (data->io.axes.z == true )
					plot->scaleRate.z = -0.003f;			
			}
			else
			{
				if (data->io.axes.x == true )
					plot->scaleRate.x = 0.003f;
				if (data->io.axes.y == true )
					plot->scaleRate.y = 0.003f;
				if (data->io.axes.z == true )
					plot->scaleRate.z = 0.003f;			
			}
			break;
		case kGraphZoomOff :
			if (data->map.plotIndex > 2)
				plot = data->map.plotChildPtr;	// get the selected branch
			plot->scaleRate.x = 0.0f;
			plot->scaleRate.y = 0.0f; 
			plot->scaleRate.z = 0.0f; 
			break;

		default : break;
	}
}

//------------------------------------------------------------------------------
void npCtrlSelect (int command, void* dataRef)
{
	int i = 0;

	pData data = (pData) dataRef;

	NPplotPtr plot = NULL;
	NPplotPtr plotChild = NULL;
	
	if (data->map.plotIndex < 3)
		plot = data->map.plot[data->map.plotIndex];
	else
		plot = data->map.plotChildPtr;


	switch (command)
	{							// make sure it copies current plot parameters, debug zz
		case kGraphNew : 
			if (plot->branchLevel > 0)
				npPlotNewChild (plot, dataRef); 
			else
				npPlotNew (kPlotPin, dataRef); 
			break;

		case kGraphDelete : npPlotDelete (plot, dataRef); break;
			
		case kGraphNext :
			if (data->map.branchLevel < 1)
			{
				if  (data->io.key.modShift)
					data->map.plotIndex--;
				else
				{
					data->map.plotIndex++;
		
					if (data->map.plotIndex >= data->map.plotCount)
						data->map.plotIndex = 3;
				}
				if (data->map.plotIndex < 3)			// don't select camera or grid
					data->map.plotIndex = data->map.plotCount - 1;

				data->map.plotChildPtr = data->map.plot[data->map.plotIndex];
			}
			else
			{	
				if (plot->branchLevel > 0)
				{
					plotChild = plot;
					plot = plotChild->parent;

					// add ability to switch pins and stay at the same level
					// traverse the branches to the the selected branch level
/*					for (i=0; i < data->map.branchLevel; i++)
						if (plotChild->childCount > plotChild->childIndex)
							plotChild = plotChild->child[plotChild->childIndex];

					plot = plotChild->parent;		// may need to cast this... debug, zz
				
					// if selected plot tree is not as deep then just use the selected leaf
					if (i < data->map.branchLevel)
*/
					if  (data->io.key.modShift)
						plot->childIndex--;
					else
						plot->childIndex++;
		
					// keep in bounds and roll over between start and end
					if (plot->childIndex >= plot->childCount)
						plot->childIndex = 0;

					if (plot->childIndex < 0)
						plot->childIndex = plot->childCount - 1;
				
					// set the ptr for easy access elsewhere, ie engine and draw
					data->map.plotChildPtr = plot->child[plot->childIndex];
				}
			}
			break;

		case kGraphNextBranch : 
			if (data->io.key.modShift)
			{
				if (plot->branchLevel > 0)		//do nothing if at the trunk
				{
					// traverse up 1 level to the childs parent
					plot = plot->parent;

					// set the global ptr and level
					data->map.plotChildPtr = plot;
					data->map.branchLevel = plot->branchLevel;
				}
			}
			else
			{	
				// traverse down 1 level of the selected child
				if (plot->childCount > plot->childIndex)
					plot = plot->child[plot->childIndex];

				// set the global ptr and level
				data->map.plotChildPtr = plot;
				data->map.branchLevel = plot->branchLevel;
			}
			break;
				
		case kGraphGrid :
			data->map.plotIndex = 2; 
			data->map.plotChildPtr = data->map.plot[data->map.plotIndex];
			data->map.branchLevel = 0;
			break;
		case kGraphCamera :
			data->map.plotIndex = 1; 
			data->map.plotChildPtr = data->map.plot[data->map.plotIndex];
			data->map.branchLevel = 0;
			break;

		case kGraphSelectAxes :
			if (data->io.axes.x == true 
				&& data->io.axes.y == true
				&& data->io.axes.z == true )
			{
				data->io.axes.z = false;
			}
			else if (data->io.axes.x == true 
					 && data->io.axes.y == true
					 && data->io.axes.z == false )
			{
				data->io.axes.x = false; 
				data->io.axes.y = false; 
				data->io.axes.z = true; 
			}
			else
			{
				data->io.axes.x = true; 
				data->io.axes.y = true; 
				data->io.axes.z = true; 
			}
			break;

		case kGraphSelect :
			plot->selected = 1 - plot->selected;
			break;
		case kGraphSelectOne :
			break;
		case kGraphSelectTwo :
			break;
		case kGraphSelectThree :
			break;
		case kGraphSelectAll :
			break;

		default : break;
	}
}


//------------------------------------------------------------------------------
void npCtrlProperty (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	NPplotPtr plot = NULL;
	NPgridPtr grid = NULL;	

	if (data->map.plotIndex < 3)
		plot = data->map.plot[data->map.plotIndex];
	else
		plot = data->map.plotChildPtr;


	switch (command)
	{
		case kGraphChannelUp :

			if (data->io.axes.x)
			{
				plot->channel.x++; 
				if (plot->channel.x >= data->io.channelCount)
					plot->channel.x = 0;
			}
			if (data->io.axes.y)
			{
				plot->channel.y++; 
				if (plot->channel.y >= data->io.channelCount)
					plot->channel.y = 0;
			}
			if (data->io.axes.z)
			{
				plot->channel.z++; 
				if (plot->channel.z >= data->io.channelCount)
					plot->channel.z = 0;
			}
			break;
		case kGraphChannelDown :
			if (data->io.axes.x)
			{
				plot->channel.x--; 
				if (plot->channel.x < 0)
					plot->channel.x = data->io.channelCount - 1;
			}
			if (data->io.axes.y)
			{
				plot->channel.y--; 
				if (plot->channel.y < 0)
					plot->channel.y = data->io.channelCount - 1;
			}
			if (data->io.axes.z)
			{
				plot->channel.z--; 
				if (plot->channel.z < 0)
					plot->channel.z = data->io.channelCount - 1;
			}
			break;

//		case kGraphSampleInterval : ; break;

		case kGraphColorUp :
			plot->colorIndex++;
			SetIndexColor (&plot->color, &plot->colorIndex);
			break;
		case kGraphColorDown :
			plot->colorIndex--;
			SetIndexColor (&plot->color, &plot->colorIndex);
			break;

		case kGraphAlphaUp :
			if (plot->color.a < 246)
				plot->color.a += 10;
			break;
		case kGraphAlphaDown :
			if (plot->color.a > 9)
				plot->color.a -= 10;
			break;

		case kGraphCenter : plot->center = 1 - plot->center; break;

		case kGraphLines :
			plot->lineWidth += 1.0f;
			if (plot->lineWidth > 5.0f)
				plot->lineWidth = 0.0f;
			break;
		case kGraphPoints :
			plot->pointSize += 1.0f;
			if (plot->pointSize > 5.0f)
				plot->pointSize = 0;
			break;
		case kGraphSurface :
			plot->surfaceType++;
			if (plot->surfaceType >= kSurfaceTypeCount)
				plot->surfaceType = 0;
			break;
		case kGraphMesh : plot->surfaceType = kSurfaceMesh; break;
		case kGraphScroll : plot->scroll = 1 - plot->scroll; break;
		case kGraphFreeze : plot->freeze = 1 - plot->freeze; break;
		case kGraphHide : plot->hide = 1 - plot->hide; break;

		case kGraphClear :
			plot->triggerOnLow.x = false;
			plot->triggerOnLow.y = false;
			plot->triggerOnLow.z = false;
			plot->triggerOnHi.x = false;
			plot->triggerOnHi.y = false;
			plot->triggerOnHi.z = false;
			break;

		case kGraphSampleInterval : 
			if (plot->type != kPlotGrid)
			{
/*				if (data->io.key.modShift)
					plot->sampleInterval /= 2;
				else
					plot->sampleInterval *= 2;
				if (plot->sampleInterval < 1)
					plot->sampleInterval = 1;
*/			}
			else
			{
				grid = (NPgridPtr) plot->data;

				if (data->io.key.modShift)
				{
					if (data->io.axes.x)
					{
						grid->segments.x--;
						if (grid->segments.x < 0)
							grid->segments.x = 0;
					}
					if (data->io.axes.y)
					{
						grid->segments.y--;
						if (grid->segments.y < 0)
							grid->segments.y = 0;
					}
					if (data->io.axes.z)
					{
						grid->segments.z--;
						if (grid->segments.z < 0)
							grid->segments.z = 0;
					}
				}
				else
				{
					if (data->io.axes.x)
						grid->segments.x++;
					if (data->io.axes.y)
						grid->segments.y++;
					if (data->io.axes.z)
						grid->segments.z++;
				}
			}
			break;

		case kEventLowSetpoint :
			if (data->io.axes.x)
			{
				plot->triggerLevelLow.x = plot->translate.x;
				plot->triggerOnLow.x = 1 - plot->triggerOnLow.x;
			}
			if (data->io.axes.y)
			{
				plot->triggerLevelLow.y = plot->translate.y;
				plot->triggerOnLow.y = 1 - plot->triggerOnLow.y;
			}
			if (data->io.axes.z)
			{
				plot->triggerLevelLow.z = plot->translate.z;
				plot->triggerOnLow.z = 1 - plot->triggerOnLow.z;
			}
			break;
		case kEventHighSetpoint :
			if (data->io.axes.x)
			{
				plot->triggerLevelHi.x = plot->translate.x;
				plot->triggerOnHi.x = 1 - plot->triggerOnHi.x;
			}
			if (data->io.axes.y)
			{
				plot->triggerLevelHi.y = plot->translate.y;
				plot->triggerOnHi.y = 1 - plot->triggerOnHi.y;
			}
			if (data->io.axes.z)
			{
				plot->triggerLevelHi.z = plot->translate.z;
				plot->triggerOnHi.z = 1 - plot->triggerOnHi.z;
			}
			break;

		case kGraphDrawIndicators :
			if (data->io.axes.x)
				plot->drawCurrentValue.x;
			if (data->io.axes.y)
				plot->drawCurrentValue.y;
			if (data->io.axes.z)
				plot->drawCurrentValue.z;
			break;

		default : break;
	}
}


//------------------------------------------------------------------------------
void npCtrlGlobal (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	NPplotPtr plot = data->map.plot[data->map.plotIndex];

	switch (command)
	{
		case kGraphMenu : data->io.drawMenu = 1 - data->io.drawMenu; break;

		case kFullScreenExit : npGlutFullScreenExit(); break;

		case kGraphClearFrameBuffer :
			data->io.clearFrameBuffer = 1 - data->io.clearFrameBuffer;
			break;

		default : break;
	}
}


//------------------------------------------------------------------------------
void npCtrlUpdate (void* dataRef)
{
	npDataUpdate (dataRef);			//update our input data, txt, video...
	npEngineUpdate (dataRef);		//position, physics, interactions...
}




